#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides the base
:class:`taurus.qt.qtgui.application.TaurusApplication` class."""

from builtins import str

import os
import sys
import logging
import threading

from taurus.external.qt import Qt

from taurus.core.util.log import LogExceptHook, Logger
from taurus import tauruscustomsettings


__all__ = ["TaurusApplication"]

__docformat__ = 'restructuredtext'


class STD(Logger):

    FlushWaterMark = 1000

    def __init__(self, name='', parent=None, format=None, std=None,
                 pure_text=True):
        """The STD Logger constructor

        :param name: (str) the logger name (default is empty string)
        :param parent: (Logger) the parent logger or None if no parent exists
                       (default is None)
        :param format: (str) the log message format or None to use the default
                       log format (default is None)
        :param std: std to forward write
        :param pure_text: if True, writes the 'message' parameter of the log
                          message in a separate line preserving the
                          indentation
        """
        Logger.__init__(self, name=name, parent=parent, format=format)
        self.buffer = ''
        self.log_obj.propagate = False
        self.std = std
    
    #Trying to mimic stderr
    @property
    def errors(self):
        return self.std.errors

    def addLogHandler(self, handler):
        """When called, set to use a private handler and DON'T send messages
        to parent loggers (basically will act as an independent logging system
        by itself)

        :param handler: new handler"""
        Logger.addLogHandler(self, handler)
        self.log_obj.propagate = not len(self.log_handlers)

    def write(self, msg):
        try:
            self.buffer += msg
            # while there is no new line, just accumulate the buffer
            msgl = len(msg)
            if msgl > 0 and \
               (msg[-1] == '\n' or msg.index('\n') >= 0 or
                    msgl >= self.FlushWaterMark):
                self.flush()
        except ValueError:
            pass
        finally:
            if self.std is not None:
                try:
                    self.std.write(msg)
                except:
                    pass
            pass

    def flush(self):
        try:
            buff = self.buffer
            if buff is None or len(buff) == 0:
                return
            # take the '\n' because the output is a list of strings, each to be
            # interpreted as a separate line in the client
            if buff[-1] == '\n':
                buff = buff[:-1]
            if self.log_handlers:
                self.log(Logger.Console, '\n' + buff)
            self.buffer = ""
        finally:
            if self.std is not None:
                try:
                    self.std.flush()
                except:
                    pass
            pass


class TaurusApplication(Qt.QApplication, Logger):
    """A QApplication that performs some taurus-specific initializations
    and (optionally but deprecated) also parses the command line for taurus
    options.

    The optional keyword parameters:
        - app_name: (str) application name
        - app_version: (str) application version
        - org_name: (str) organization name
        - org_domain: (str) organization domain
        - cmd_line_parser (None [or DEPRECATED :class:`optparse.OptionParser`])

    If cmd_line_parser is explicitly set to None, no parsing will
    be done at all. If a  :class:`optparse.OptionParser` instance is passed as
    cmd_line_parser (deprecated), it will be used for parsing the command line
    arguments. If cmd_line_parser is not explicitly passed, the behaviour will
    depend on the value of tauruscustomsettings.IMPLICIT_OPTPARSE (which by
    default is False, indicating that no parsing is done).

    Simple example::

        import sys
        from taurus.qt.qtgui.application import TaurusApplication
        import taurus.qt.qtgui.display

        app = TaurusApplication(cmd_line_parser=None)

        w = taurus.qt.qtgui.display.TaurusLabel()
        w.model = 'sys/tg_test/1/double_scalar'
        w.show()

        sys.exit(app.exec_())

    """

    def __init__(self, *args, **kwargs):
        """The constructor. Parameters are the same as QApplication plus a
        keyword parameter: 'cmd_line_parser' which should be None or an
        instance of :class:`optparse.OptionParser`. If cmd_line_parser is None
        no command line parsing will be done ()
        """

        # lock to safely get singleton elements (like IPython taurus
        # console app)
        self._lock = threading.Lock()

        if len(args) == 0:
            args = getattr(sys, 'argv', []),

        app_name = kwargs.pop('app_name', None)
        app_version = kwargs.pop('app_version', None)
        org_name = kwargs.pop('org_name', None)
        org_domain = kwargs.pop('org_domain', None)

        if getattr(tauruscustomsettings, 'IMPLICIT_OPTPARSE', True):
            import optparse
            default_parser = optparse.OptionParser()
        else:
            default_parser = None
        parser = kwargs.pop('cmd_line_parser', default_parser)

        #######################################################################
        # Workaround for XInitThreads-related crash.
        # Sometimes (e.g. running `taurusgui example01`) one gets the following:
        # [xcb] Unknown request in queue while dequeuing
        # [xcb] Most likely this is a multi-threaded client and XInitThreads has not been called
        # [xcb] Aborting, sorry about that.
        #
        # According to http://stackoverflow.com/a/31967769 , it is fixed by:
        try:
            Qt.QCoreApplication.setAttribute(Qt.Qt.AA_X11InitThreads)
        except AttributeError:
            pass  # Qt < 4.8 does not define AA_X11InitThreads
        ######################################################################

        try:
            Qt.QApplication.__init__(self, *args, **kwargs)
        except TypeError:
            Qt.QApplication.__init__(self, *args)

        Logger.__init__(self)

        self._out = None
        self._err = None

        if app_name is not None:
            self.setApplicationName(app_name)
        if app_version is not None:
            self.setApplicationVersion(app_version)
        if org_name is not None:
            self.setOrganizationName(org_name)
        if org_domain is not None:
            self.setOrganizationDomain(org_domain)

        if parser is None:
            p, opt, args = None, None, None
        else:
            try:
                if parser.version is None and app_version:
                    # if the parser does not contain version information
                    # but we have an application version, add the
                    # --version capability
                    v = app_version
                    if app_name:
                        v = app_name + " " + app_version
                    parser.version = v
                    parser._add_version_option()

                import taurus.core.util.argparse
                p, opt, args = taurus.core.util.argparse.init_taurus_args(
                    parser=parser, args=args[0][1:])
            except Exception:
                self.error("Error while using the given cmd_line_parser. \n"
                           + "hint: it must be either None or an "
                           + "optparse.OptionParser instance")
                raise

        self._cmd_line_parser = p
        self._cmd_line_options = opt
        self._cmd_line_args = args
        self.__registerQtLogger()
        self.__registerExtensions()
        self.__redirect_std()

    def __registerQtLogger(self):
        import taurus.qt.qtcore.util
        taurus.qt.qtcore.util.initTaurusQtLogger()

    def __registerExtensions(self):
        """Registers taurus Qt extensions"""
        try:
            import sardana.taurus.qt.qtcore.tango.sardana
            sardana.taurus.qt.qtcore.tango.sardana.registerExtensions()
        except:
            self.debug("Failed to load sardana extensions", exc_info=1)
        try:
            import taurus.core.tango.img
            taurus.core.tango.img.registerExtensions()
        except:
            self.debug("Failed to load image extensions", exc_info=1)

    def __redirect_std(self):
        """Internal method to redirect stdout and stderr to log messages"""
        Logger.addLevelName(Logger.Critical + 10, 'CONSOLE')
        # only redirect if display hook has not been set (IPython does it)
        if sys.displayhook == sys.__displayhook__:
            self._out = STD(name="OUT", std=sys.stdout)
            sys.stdout = self._out
            self._err = STD(name="ERR", std=sys.stderr)
            sys.stderr = self._err

    def __buildLogFileName(self, prefix=None, name=None):
        if prefix is None:
            prefix = os.path.expanduser('~/tmp')
        appName = str(self.applicationName())
        if not appName:
            appName = os.path.splitext(os.path.basename(sys.argv[0]))[0]
        dirName = os.path.join(prefix, appName)
        if not os.path.isdir(dirName):
            os.makedirs(dirName)
        if name is None:
            name = appName + '.log'
        fileName = os.path.join(dirName, name)
        return fileName

    def get_command_line_parser(self):
        """Returns the :class:`optparse.OptionParser` used to parse the
        command line parameters.

        :return: the parser used in the command line
        :rtype: :class:`optparse.OptionParser`"""
        # TODO: issue a warning if this is uninitialized (cmd_line_parser=None)
        return self._cmd_line_parser

    def get_command_line_options(self):
        """Returns the :class:`optparse.Option` that resulted from parsing the
        command line parameters.

        :return: the command line options
        :rtype: :class:`optparse.Option`"""
        # TODO: issue a warning if this is uninitialized (cmd_line_parser=None)
        return self._cmd_line_options

    def get_command_line_args(self):
        """Returns the list of arguments that resulted from parsing the
        command line parameters.

        :return: the command line arguments
        :rtype: list of strings"""
        # TODO: issue a warning if this is uninitialized (cmd_line_parser=None)
        return self._cmd_line_args

    def setTaurusStyle(self, styleName):
        """Sets taurus application style to the given style name

        :param styleName: the new style name to be applied
        :type styleName: str"""
        import taurus.qt.qtgui.style
        taurus.qt.qtgui.style.setTaurusStyle(styleName)

    def basicConfig(self, log_file_name=None, maxBytes=1E7, backupCount=5,
                    with_gui_exc_handler=True):

        # prepare exception handler to report exceptions as error log messages
        # and to taurus message box (if with_gui is set)
        hook_to = None
        if with_gui_exc_handler:
            import taurus.qt.qtgui.dialog
            hook_to = taurus.qt.qtgui.dialog.TaurusExceptHookMessageBox()
        sys.excepthook = LogExceptHook(hook_to=hook_to)

        # create a file log handler
        try:
            if log_file_name is None:
                log_file_name = self.__buildLogFileName()
            f_h = logging.handlers.RotatingFileHandler(log_file_name,
                                                       maxBytes=maxBytes, backupCount=backupCount)
            Logger.addRootLogHandler(f_h)
            if self._out is not None:
                self._out.std = sys.__stdout__
                self._out.addLogHandler(f_h)
            if self._out is not None:
                self._err.std = sys.__stderr__
                self._err.addLogHandler(f_h)
            self.info("Logs will be saved in %s", log_file_name)
        except:
            self.warning("'%s' could not be created. Logs will not be stored",
                         log_file_name)
            self.debug("Error description", exc_info=1)

    @staticmethod
    def exec_(*args, **kwargs):
        # TODO: substitute this ugly hack by some more general mechanism
        try:
            ret = Qt.QApplication.exec_(*args, **kwargs)
        except TypeError:
            ret = Qt.QApplication.exec_(*args)
        from taurus.core.util.log import _DEPRECATION_COUNT
        from taurus import info
        info('\n*********************\n%s', _DEPRECATION_COUNT.pretty())
        return ret



# -*- coding: utf-8 -*-

##############################################################################
##
## This file is part of Taurus
##
## http://taurus-scada.org
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module exposes PyQt4/PyQt5/PySide module"""

__all__ = ["initialize", "API_NAME",
           "getQtName", "getQt", "requires"]

import os
import sys
import platform
from distutils.version import LooseVersion
from taurus.core.util import log as __log
from taurus import tauruscustomsettings as __config


class PythonQtError(RuntimeError):
    """Error raised if no bindings could be selected."""
    pass


#: Qt API environment variable name
QT_API = 'QT_API'
#: names of the PyQt5 api
PYQT5_API = ['pyqt5']
#: names of the PyQt4 api
PYQT4_API = [
    'pyqt',  # name used in IPython.qt
    'pyqt4'  # pyqode.qt original name
]
#: names of the PySide api
PYSIDE_API = ['pyside']

#: names of the PySide2 api
PYSIDE2_API = ['pyside2']

#: The constants PYQT5, PYQT4, PYSIDE, PYSIDE2, PYSIDE_VERSION and PYQT_VERSION
#: will be updated depending on the selected binding
PYQT5 = PYQT4 = PYSIDE = PYSIDE2 = False
PYSIDE_VERSION = PYQT_VERSION = None

# First, check if some binding is already in use (and, if so, select it)
if 'PyQt5' in sys.modules:
    API = 'pyqt5'
elif 'PySide2' in sys.modules:
    API = 'pyside2'
elif 'PyQt4' in sys.modules:
    API = 'pyqt4'
elif 'PySide' in sys.modules:
    API = 'pyside'
else:
    # if no binding is already loaded, use (in this order):
    #   - QT_API environment variable
    #   - tauruscustomsettings.DEFAULT_QT_API
    #   - first successful import of 'pyqt5', 'pyqt4', 'pyside2', 'pyside'
    API = os.environ.get(QT_API, getattr(__config, 'DEFAULT_QT_API', ''))
    API = API.lower()


if API not in (PYQT5_API + PYQT4_API + PYSIDE_API + PYSIDE2_API + ['']):
    raise ImportError("Unknown Qt API '{}'".format(API))


if not API or API in PYQT5_API:
    try:
        from PyQt5.QtCore import PYQT_VERSION_STR as PYQT_VERSION
        from PyQt5.QtCore import QT_VERSION_STR as QT_VERSION

        PYSIDE_VERSION = None
        PYQT4 = PYSIDE = PYSIDE2 = False
        PYQT5 = True
        API = os.environ['QT_API'] = 'pyqt5'

        if sys.platform == 'darwin':
            macos_version = LooseVersion(platform.mac_ver()[0])
            if macos_version < LooseVersion('10.10'):
                if LooseVersion(QT_VERSION) >= LooseVersion('5.9'):
                    raise PythonQtError("Qt 5.9 or higher only works in "
                                       + "macOS 10.10 or higher. Your "
                                       + "program will fail in this "
                                       + "system.")
            elif macos_version < LooseVersion('10.11'):
                if LooseVersion(QT_VERSION) >= LooseVersion('5.11'):
                    raise PythonQtError("Qt 5.11 or higher only works in "
                                       + "macOS 10.11 or higher. Your "
                                       + "program will fail in this "
                                       + "system.")

            del macos_version
    except ImportError:
        if API:  # if an specific API was requested, fail with import error
            raise ImportError('Cannot import PyQt5')
        # if no specific API was requested, allow trying other bindings
        __log.debug('Cannot import PyQt5')

if not API or API in PYQT4_API:
    try:
        import sip

        sip.setapi('QString', 2)
        sip.setapi('QVariant', 2)
        sip.setapi('QDate', 2)
        sip.setapi('QDateTime', 2)
        sip.setapi('QTextStream', 2)
        sip.setapi('QTime', 2)
        sip.setapi('QUrl', 2)
        from PyQt4.QtCore import PYQT_VERSION_STR as PYQT_VERSION  # analysis:ignore
        from PyQt4.QtCore import QT_VERSION_STR as QT_VERSION  # analysis:ignore

        PYSIDE_VERSION = None
        PYQT5 = PYSIDE = PYSIDE2 = False
        PYQT4 = True
        API = os.environ['QT_API'] = 'pyqt'  # "pyqt4" is forced to "pyqt"
    except ImportError:
        if API:  # if an specific API was requested, fail with import error
            raise ImportError('Cannot import PyQt4')
        # if no specific API was requested, allow trying other bindings
        __log.debug('Cannot import PyQt4')

if not API or API in PYSIDE2_API:
    try:
        from PySide2 import __version__ as PYSIDE_VERSION  # analysis:ignore
        from PySide2.QtCore import __version__ as QT_VERSION  # analysis:ignore

        PYQT_VERSION = None
        PYQT5 = PYQT4 = PYSIDE = False
        PYSIDE2 = True
        API = os.environ['QT_API'] = 'pyside2'

        if sys.platform == 'darwin':
            macos_version = LooseVersion(platform.mac_ver()[0])
            if macos_version < LooseVersion('10.11'):
                if LooseVersion(QT_VERSION) >= LooseVersion('5.11'):
                    raise PythonQtError("Qt 5.11 or higher only works in "
                                       + "macOS 10.11 or higher. Your "
                                       + "program will fail in this "
                                       + "system.")

            del macos_version
    except ImportError:
        if API:  # if an specific API was requested, fail with import error
            raise ImportError('Cannot import PySide2')
        # if no specific API was requested, allow trying other bindings
        __log.debug('Cannot import PySide2')

if not API or API in PYSIDE_API:
    try:
        from PySide import __version__ as PYSIDE_VERSION  # analysis:ignore
        from PySide.QtCore import __version__ as QT_VERSION  # analysis:ignore

        PYQT_VERSION = None
        PYQT5 = PYQT4 = PYSIDE = False
        PYSIDE = True
        API = os.environ['QT_API'] = 'pyside'

    except ImportError:
        if API:  # if an specific API was requested, fail with import error
            raise ImportError('Cannot import PySide')
        # if no specific API was requested, allow trying other bindings
        __log.debug('Cannot import PySide')

if not API:
    raise ImportError('No Qt bindings could be imported')

API_NAME = {'pyqt5': 'PyQt5', 'pyqt': 'PyQt4', 'pyqt4': 'PyQt4',
            'pyside': 'PySide', 'pyside2': 'PySide2'}[API]

# Update the environment so that other libraries that also use the same
# convention (such as guidata or spyder) do a consistent choice
os.environ['QT_API'] = API


def __initializeQtLogging():
    from importlib import import_module
    QtCore = import_module(API_NAME + '.QtCore')

    QT_LEVEL_MATCHER = {
        QtCore.QtDebugMsg: __log.debug,
        QtCore.QtWarningMsg: __log.warning,
        QtCore.QtCriticalMsg: __log.critical,
        QtCore.QtFatalMsg: __log.fatal,
        QtCore.QtSystemMsg: __log.critical,
    }

    if hasattr(QtCore, "qInstallMessageHandler"):
        # Qt5
        def taurusMessageHandler(msg_type, log_ctx, msg):
            f = QT_LEVEL_MATCHER.get(msg_type)
            return f("Qt%s %s.%s[%s]: %s", log_ctx.category, log_ctx.file,
                     log_ctx.function, log_ctx.line, msg)

        QtCore.qInstallMessageHandler(taurusMessageHandler)
    elif hasattr(QtCore, "qInstallMsgHandler"):
        # Qt4
        def taurusMsgHandler(msg_type, msg):
            f = QT_LEVEL_MATCHER.get(msg_type)
            return f("Qt: " + msg)

        QtCore.qInstallMsgHandler(taurusMsgHandler)


def __removePyQtInputHook():
    from importlib import import_module
    QtCore = import_module(API_NAME + '.QtCore')
    if hasattr(QtCore, "pyqtRemoveInputHook"):
        QtCore.pyqtRemoveInputHook()


def __addExceptHook():
    """
    Since PyQt 5.5 , unhandled python exceptions cause the application to
    abort:

    http://pyqt.sf.net/Docs/PyQt5/incompatibilities.html#unhandled-python-exceptions

    By calling __addExceptHook, we restore the old behaviour (just print the
    exception trace).
    """
    import traceback
    sys.excepthook = traceback.print_exception


if getattr(__config, 'QT_AUTO_INIT_LOG', True):
    __initializeQtLogging()

if getattr(__config, 'QT_AUTO_REMOVE_INPUTHOOK', True):
    __removePyQtInputHook()

if PYQT5 and getattr(__config, 'QT_AVOID_ABORT_ON_EXCEPTION', True):
    # TODO: check if we also want to do this for PySide(2)
    __addExceptHook()

__log.info('Using %s (v%s with Qt %s and Python %s)',
           API_NAME,
           PYQT_VERSION or PYSIDE_VERSION,
           QT_VERSION,
           sys.version.split()[0])


# --------------------------------------------------------------------------
# Deprecated (in Jul17) pending to be removed later on


def getQt(name=None, strict=True):
    __log.deprecated(dep='taurus.external.qt.getQt', rel='4.0.4')
    from importlib import import_module
    return import_module(API_NAME)


def getQtName(name=None, strict=True):
    __log.deprecated(dep='taurus.external.qt.getQtName',
                     alt='taurus.external.qt.API_NAME', rel='4.0.4')
    return API_NAME


def initialize(name=None, strict=True, logging=True,
               resources=True, remove_inputhook=True):
    __log.deprecated(dep='taurus.external.qt.initialize', rel='4.0.4')
    return getQt()


def requires(origin=None, exc_class=ImportError, **kwargs):
    __log.deprecated(dep='taurus.external.qt.requires', rel='4.0.4')
    return True


# Handle rename of DEFAULT_QT_AUTO_API -->  DEFAULT_QT_API
if hasattr(__config, 'DEFAULT_QT_AUTO_API'):
    __log.deprecated(dep='DEFAULT_QT_AUTO_API', alt='DEFAULT_QT_API',
                     rel='4.0.4')
    if not hasattr(__config, 'DEFAULT_QT_API'):
        __config.DEFAULT_QT_API = __config.DEFAULT_QT_AUTO_API

# --------------------------------------------------------------------------



#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
This module provides the taurus Command Line Interface

It is based on the click module to provide CLI utilities.

It defines a `taurus` command which can accept subcommands defined in
other taurus submodules or even in 3rd party plugins.

To define a subcommand for taurus, you need to register it in the
`taurus.cli.subcommands` entry point group via setuptools.

.. todo:: add click link , add code snippet for plugins, etc.

This module provides also common options/flags for command-line interfaces
that can be used by taurus plugins. For list of available options,
see :mod:`taurus.cli.common`

.. seealso:: :mod:`taurus.cli.alt`

"""

from .cli import main, register_subcommands, taurus_cmd



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################


"""Unit tests for TaurusCommandButton"""


import pytest
from taurus.qt.qtgui.button import TaurusCommandButton
from taurus.external.qt import Qt

try:
    # The following are Tango-centric imports.
    # TODO: change them if/when TaurusCommandbuttongets generalized
    from PyTango import CommunicationFailed
    from taurus.core.tango.test import taurus_test_ds  # pytest fixture
    _TANGO_MISSING = False
except:
    _TANGO_MISSING = True


@pytest.mark.skipif(_TANGO_MISSING, reason="tango-dependent test")
def test_timeout(qtbot, taurus_test_ds):
    """Check that the timeout property works"""
    w = TaurusCommandButton(command='Sleep')
    qtbot.addWidget(w)

    try:
        w.setModel(taurus_test_ds)

        # set the parameter for the sleep command to 0.2 seconds
        w.setParameters([.2])

        # Create a callback to check the result of the execution
        def _check_result(res):
            return res == .2

        # Check that the button emits commandExecuted when timeout > exec time
        w.setTimeout(10)
        assert w.getTimeout() == 10
        with qtbot.waitSignal(w.commandExecuted, check_params_cb=_check_result):
            qtbot.mouseClick(w, Qt.Qt.LeftButton)

        # Check that, if timeout < exec time, commandExecuted is NOT emited
        # and CommunicationFailed is raised
        w.setTimeout(.1)
        assert w.getTimeout() == .1
        with pytest.raises(CommunicationFailed):
            with qtbot.assertNotEmitted(w.commandExecuted):
                # call _onClicked instead of emulating a click to bypass
                # the @ProtectTaurusMessageBox protection of onClicked()
                w._onClicked()
    finally:
        # just in case this helps avoiding problems at exit
        w.setModel(None)







#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################


import sys
from types import ModuleType
from taurus.core.util.lazymodule import LazyModule

from pkg_resources import EntryPoint, Requirement, WorkingSet


def test_LazyModule():

    # create an entry point for taurus.core.util.test.dumm
    w = WorkingSet()
    d = w.find(Requirement.parse('taurus'))
    ep = EntryPoint.parse("dummy_mod = taurus.core.util.test.dummy", dist=d)
    modname = ep.name

    # lazy load the ep module as taurus.fbt
    LazyModule.import_ep(modname, "taurus", ep)

    # check that lazy-loading did not import the entry point modules
    assert modname not in sys.modules
    assert ep.module_name not in sys.modules

    # import the module and check that it is a LazyModule
    import taurus.dummy_mod as lzm
    assert isinstance(lzm, LazyModule)

    # same again
    import taurus.dummy_mod as lzm
    assert isinstance(lzm, LazyModule)

    # now access a member of the lazy module
    assert lzm.foo == 1

    # ...and check that any subsequent import will return a "real" module,
    # not a lazy one

    import taurus.dummy_mod as lzm
    assert not isinstance(lzm, LazyModule)
    assert isinstance(lzm, ModuleType)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Common tools and fixtures for using with taurus and pytest"""

import pytest
import contextlib
from taurus.external.qt import PYSIDE2


@contextlib.contextmanager
def check_taurus_deprecations(caplog, expected=0):
    """context manager that checks the number of taurus deprecation warnings
    logged within the context execution.

    :param expected: (int) Expected number of deprecations. Default is 0
    """
    from taurus import tauruscustomsettings
    # disable deprecation silencing in this context
    bck = getattr(tauruscustomsettings, "MAX_DEPRECATIONS_LOGGED", None)
    tauruscustomsettings._MAX_DEPRECATIONS_LOGGED = None
    caplog.clear()
    try:
        yield
    finally:
        tauruscustomsettings._MAX_DEPRECATIONS_LOGGED = bck
    # check the deprecations after the context is run
    deps = [r for r in caplog.records if "DeprecationWarning" in r.msg]
    n = len(deps)
    msg = "{} Deprecation Warnings ({} expected)".format(n, expected)
    if PYSIDE2 and len(deps) != expected:
        # TODO: investigate the cause for this (note that it happens only
        #       if taurus.external.qt has been imported)
        pytest.xfail("log handling is not working as expected for PySide2")
    assert len(deps) == expected, msg



#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
a misc collection of basic types
"""

import datetime

from .util.enumeration import Enumeration
from .util.log import taurus4_deprecation
from enum import IntEnum
from future.utils import PY2
from builtins import object


__all__ = ["TaurusSWDevState", "TaurusSWDevHealth", "OperationMode",
           "TaurusSerializationMode", "SubscriptionState", "TaurusEventType",
           "MatchLevel", "TaurusElementType", "LockStatus", "DataFormat",
           "AttrQuality", "AttrAccess", "DisplayLevel", "ManagerState",
           "TaurusTimeVal", "TaurusAttrValue", "TaurusConfigValue", "DataType",
           "TaurusLockInfo", "TaurusDevState", "TaurusModelValue"]

__docformat__ = "restructuredtext"


class TaurusDevState(IntEnum):
    """Enumeration of possible states of :class:`taurus.core.TaurusDevice`
    objects. This is returned, e.g. by :meth:`TaurusDevice.state`.

    The description of the values of this enumeration is:

    - Ready: the device can be operated by the user and could even be
      involved in some operation.
    - NotReady: the device can not be operated by the user (e.g. due to
      still being initialized, or due to a device failure,...)
    - Undefined: it is not possible to retrieve a coherent state from the
      device (e.g. due to communication, or to contradictory internal
      states, ...)
    """
    # TODO: it could be extended for more detailed description using bit masks
    Ready = 1
    NotReady = 2
    Undefined = 4

# Deprecated enumeration.
# According with TEP3, the logic of the OperationMode should be in the widgets
OperationMode = Enumeration(
    'OperationMode', (
        'OFFLINE',
        'ONLINE'
    ))

TaurusSerializationMode = Enumeration(
    'TaurusSerializationMode', (
        'Serial',
        'Concurrent',
        'TangoSerial',
    ))

TaurusEventType = Enumeration(
    'TaurusEventType', (
        'Change',
        'Config',  # Deprecated in Taurus >=4
        'Periodic',
        'Error'
    ))

MatchLevel = Enumeration(
    'MatchLevel', (
        'ANY',
        'SHORT',
        'NORMAL',
        'COMPLETE',
        'SHORT_NORMAL',
        'NORMAL_COMPLETE'
    ))

TaurusElementType = Enumeration(
    'TaurusElementType', (
        'Unknown',
        'Name',
        'DeviceClass',
        'Device',
        'DeviceAlias',
        'Domain',
        'Family',
        'Member',
        'Server',
        'ServerName',
        'ServerInstance',
        'Exported',
        'Host',
        'Attribute',
        'AttributeAlias',
        'Command',
        'Property',
        'Configuration',
        'Authority'
    ))

LockStatus = Enumeration(
    'LockStatus', (
        'Unlocked',
        'Locked',
        'LockedMaster',
        'Unknown'
    ))

DataFormat = Enumeration(
    'DataFormat', (
        '_0D',
        '_1D',
        '_2D'
    ))

# TODO: Consider adding  Enum and Quantity to DataType enumeration ...
# TODO: ... and also to __PYTHON_TYPE_TO_TAURUS_DATATYPE

DataType = Enumeration(
    'DataType', (
        'Integer',  # Can be used in scheme-agnostic code
        'Float',  # Can be used in scheme-agnostic code
        'String',  # Can be used in scheme-agnostic code
        'Boolean',  # Can be used in scheme-agnostic code
        'Bytes',
        'DevState',  # This type is for Tango backwards-compat. Avoid using it
        'DevEncoded',  # This type is for Tango backwards-compat. Avoid using it
        'Object'  # use this for opaque py objects not described by any of the above
    ))
# monkey-patch DataType to provide from_python_type()
__PYTHON_TYPE_TO_TAURUS_DATATYPE = {
    str: DataType.String,
    int: DataType.Integer,
    # long : DataType.Integer,  # (only in py2) see below...
    float: DataType.Float,
    bool: DataType.Boolean,
    # bytes : DataType.Bytes,  # (only in py>=3) see below...
}
if PY2:  # Python 2
    __PYTHON_TYPE_TO_TAURUS_DATATYPE[long] = DataType.Integer
else:  # Python >=3
    __PYTHON_TYPE_TO_TAURUS_DATATYPE[bytes] = DataType.Bytes
DataType.from_python_type = __PYTHON_TYPE_TO_TAURUS_DATATYPE.get

SubscriptionState = Enumeration(
    "SubscriptionState", (
        "Unsubscribed",
        "Subscribing",
        "Subscribed",
        "PendingSubscribe"
    ))


class AttrQuality(IntEnum):
    """Enumeration of quality states for Taurus attributes. based on
    This is the Taurus equivalent to PyTango.AttrQuality.
    The members present in PyTango are also defined here with the same values,
    allowing equality comparisons with :class:`PyTango.AttrQuality` (but not
    identity checks!)::

        from taurus.core import AttrQuality as Q1
        from PyTango import AttrQuality as Q2

        Q1.ATTR_ALARM == Q2.ATTR_ALARM                  # --> True
        Q1.ATTR_ALARM in (Q2.ATTR_ALARM, Q2.ATTR_ALARM) # --> True
        Q1.ATTR_ALARM == 2                              # --> True
        Q1.ATTR_ALARM is 2                              # --> False
        Q1.ATTR_ALARM is Q2.ATTR_ALARM                  # --> False
    """
    ATTR_VALID = 0
    ATTR_INVALID = 1
    ATTR_ALARM = 2
    ATTR_CHANGING = 3
    ATTR_WARNING = 4

    def __str__(self):
        # Todo: BEWARE!! This alters the usual behaviour of IntEnum!
        return self.name

AttrAccess = Enumeration(
    'AttrAccess', (
        'READ',
        'READ_WITH_WRITE',
        'WRITE',
        'READ_WRITE'
    ))

DisplayLevel = Enumeration(
    'DisplayLevel', (
        'OPERATOR',
        'EXPERT',
        'DEVELOPER'
    ))

ManagerState = Enumeration(
    'ManagerState', (
        'UNINITIALIZED',
        'INITED',
        'CLEANED'
    ))


class DeprecatedEnum(object):

    def __init__(self, name, alt):
        self.__name = name
        self.__alt = alt

    def __getattr__(self, name):
        raise RuntimeError('%s enumeration was removed. Use %s instead' %
                           (self.__name, self.__alt))


TaurusSWDevState = DeprecatedEnum('TaurusSWDevState', 'TaurusDevState')
TaurusSWDevHealth = DeprecatedEnum('TaurusSWDevHealth', 'TaurusDevState')


class TaurusTimeVal(object):

    def __init__(self):
        self.tv_sec = 0
        self.tv_usec = 0
        self.tv_nsec = 0

    def __repr__(self):
        return "%s(tv_sec=%i, tv_usec=%i, tv_nsec=%i)" % (self.__class__.__name__, self.tv_sec, self.tv_usec, self.tv_nsec)

    def __float__(self):
        return self.totime()

    def totime(self):
        return self.tv_nsec * 1e-9 + self.tv_usec * 1e-6 + self.tv_sec

    def todatetime(self):
        return datetime.datetime.fromtimestamp(self.totime())

    def isoformat(self):
        return self.todatetime().isoformat()

    @staticmethod
    def fromtimestamp(v):
        tv = TaurusTimeVal()
        tv.tv_sec = int(v)
        usec = (v - tv.tv_sec) * 1000000
        tv.tv_usec = int(usec)
        tv.tv_nsec = int((usec - tv.tv_usec) * 1000)
        return tv

    @staticmethod
    def fromdatetime(v):
        import time
        tv = TaurusTimeVal()
        tv.tv_sec = int(time.mktime(v.timetuple()))
        tv.tv_usec = v.microsecond
        tv.tv_nsec = 0  # datetime does not provide ns info
        return tv

    @staticmethod
    def now():
        return TaurusTimeVal.fromdatetime(datetime.datetime.now())


class TaurusModelValue(object):

    def __init__(self):
        self.rvalue = None

    def __repr__(self):
        return "%s%s" % (self.__class__.__name__, repr(self.__dict__))


class TaurusAttrValue(TaurusModelValue):

    def __init__(self):
        TaurusModelValue.__init__(self)
        self.wvalue = None
        self.time = None
        self.quality = AttrQuality.ATTR_VALID
        self.error = None


class TaurusConfigValue(object):

    @taurus4_deprecation(alt='TaurusAttrValue')
    def __init__(self):
        pass


class TaurusLockInfo(object):

    LOCK_STATUS_UNKNOWN = 'Lock status unknown'

    def __init__(self):
        self.status = LockStatus.Unknown
        self.status_msg = self.LOCK_STATUS_UNKNOWN
        self.id = None
        self.host = None
        self.klass = None

    def __repr__(self):
        return self.status_msg


#del time
#del datetime
#del Enumeration
#del AttrQuality, AttrAccess, DataFormat, LockStatus



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides an Input panel (usually used inside a TaurusDialog)"""
from __future__ import print_function

from builtins import str
from builtins import object

try:
    from collections.abc import Sequence, Mapping
except ImportError:  # bck-compat py 2.7
    from collections import Sequence, Mapping
import numpy

from future.utils import string_types

from taurus.external.qt import Qt
from taurus.qt.qtgui.util.ui import UILoadable

__all__ = ["TaurusInputPanel"]

__docformat__ = 'restructuredtext'


@UILoadable(with_ui='_ui')
class TaurusInputPanel(Qt.QWidget):
    """A panel design to get an input from the user.

    The input_data is a dictionary which contains information on how to build
    the input dialog. It **must** contains the following keys:

        - *prompt* <str>: message to be displayed

    The following are optional keys (and their corresponding default values):

        - *title* <str> (doesn't have default value)
        - *key* <str> (doesn't have default value):
          a label to be presented left to the input box represeting the label
        - *unit* <str> (doesn't have default value):
          a label to be presented right to the input box representing the units
        - *data_type* <str or sequence> ('String'):
          type of data to be requested. Standard
          accepted data types are 'String', 'Integer', 'Float', 'Boolean',
          'Text'. A list of elements will be interpreted as a selection.
          Default TaurusInputPanel class will interpret any custom data types as
          'String' and will display input widget accordingly. Custom
          data types can be handled differently by supplying a different
          input_panel_klass.
        - *minimum* <int/float>:
          minimum value (makes sense when data_type is 'Integer' or 'Float')
        - *maximum* <int/float>:
          maximum value (makes sense when data_type is 'Integer' or 'Float')
        - *step* <int/float> (1):
          step size value (makes sense when data_type is 'Integer' or 'Float')
        - *decimals* <int> (1):
          number of decimal places to show (makes sense when data_type is
          'Float')
        - *default_value* <obj> (doesn't have default value):
          default value
        - *allow_multiple* <bool> (False):
          allow more than one value to be selected (makes sense when data_type
          is a sequence of possibilities)


    Example::

        app = Qt.QApplication([])

        class Listener(object):
            def on_accept(self):
                print "user selected", self.panel.value()

        d = dict(prompt="What's your favourite car brand?",
                 data_type=["Mazda", "Skoda", "Citroen", "Mercedes", "Audi", "Ferrari"],
                 default_value="Mercedes")
        w = TaurusInputPanel(d)
        l = Listener()
        l.panel = w
        w.connect(w.buttonBox(), Qt.SIGNAL("accepted()"), l.on_accept)
        w.show()
        app.exec_()
    """

    def __init__(self, input_data, parent=None):
        Qt.QWidget.__init__(self, parent)
        self._input_data = input_data
        self.loadUi()
        self.fill_main_panel(self._ui._inputPanel, input_data)

    def fill_main_panel(self, panel, input_data):
        layout = Qt.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        panel.setLayout(layout)
        if isinstance(input_data, Mapping):
            single_panel, getter = self.create_single_input_panel(input_data)
            layout.addWidget(single_panel)
            self.value = getter
            if 'title' in input_data:
                self.setWindowTitle(input_data['title'])

    def create_single_input_panel(self, input_data):
        style = Qt.QApplication.instance().style()
        icon = style.standardIcon(Qt.QStyle.SP_MessageBoxQuestion)
        self.setIconPixmap(icon.pixmap(64))

        self.setText(input_data['prompt'])

        data_type = input_data.get('data_type', 'String')
        is_seq = not isinstance(data_type, string_types) and \
            isinstance(data_type, Sequence)
        if is_seq:
            panel, getter = self.create_selection_panel(input_data)
        else:
            data_type_l = data_type.lower()
            creator = getattr(self, "create_" + data_type_l +
                              "_panel", self.create_custom_panel)
            if creator:
                panel, getter = creator(input_data)

        if panel is None:
            panel = Qt.QLabel(
                "Cannot create widget for data type '%s'" % data_type)
            getter = lambda: None
        return panel, getter

    def create_custom_panel(self, input_data):
        return self.create_string_panel(input_data)

    def create_selection_panel(self, input_data):
        allow_multiple = input_data.get('allow_multiple', False)

        if allow_multiple:
            return self._create_multi_selection_panel(input_data)
        else:
            return self._create_single_selection_panel(input_data)

    def _create_single_selection_panel(self, input_data):
        items = list(map(str, input_data['data_type']))
        if len(items) > 5:
            return self._create_combobox_panel(input_data)
        return self._create_radiobutton_panel(input_data)

    def _create_combobox_panel(self, input_data):
        panel = self._create_simple_panel(input_data)
        layout = panel.layout()
        self._ui.inputWidget = combobox = Qt.QComboBox()
        items = input_data['data_type']
        for item in items:
            is_seq = not isinstance(item, string_types) and \
                isinstance(item, Sequence)
            if is_seq:
                text, userData = item
            else:
                text, userData = str(item), item
            combobox.addItem(text, userData)
        layout.addWidget(combobox, 0, 1)
        return panel, self._get_combobox_value

    def _get_combobox_value(self):
        combo = self._ui.inputWidget
        return combo.itemData(combo.currentIndex())

    def _create_radiobutton_panel(self, input_data):
        panel = self._create_group_panel(input_data)
        layout = panel.layout()
        items = input_data['data_type']
        default_value = input_data.get('default_value')
        self._ui.inputWidget = buttongroup = Qt.QButtonGroup()
        buttongroup.setExclusive(True)
        for item in items:
            is_seq = not isinstance(item, string_types) and \
                isinstance(item, Sequence)
            if is_seq:
                text, userData = item
            else:
                text, userData = str(item), item
            button = Qt.QRadioButton(text)
            button._value = userData
            if default_value == userData:
                button.setChecked(True)
            buttongroup.addButton(button)
            layout.addWidget(button)
        return panel, self._get_radiobutton_value

    def _get_radiobutton_value(self):
        buttongroup = self._ui.inputWidget
        button = buttongroup.checkedButton()
        if button is not None:
            return button._value

    def _create_multi_selection_panel(self, input_data):
        panel = self._create_group_panel(input_data)
        layout = panel.layout()
        items = input_data['data_type']
        default_value = input_data.get('default_value')
        if default_value is None:
            default_value = ()
        dft_is_seq = not isinstance(default_value, string_types) and \
            isinstance(default_value, Sequence)
        if not dft_is_seq:
            default_value = default_value,

        self._ui.inputWidget = listwidget = Qt.QListWidget()
        listwidget.setSelectionMode(Qt.QAbstractItemView.MultiSelection)

        for item in items:
            is_seq = not isinstance(item, string_types) and \
                isinstance(item, Sequence)
            if is_seq:
                text, userData = item
            else:
                text, userData = str(item), item
            item_widget = Qt.QListWidgetItem(text, listwidget)
            item_widget.setData(Qt.Qt.UserRole, userData)
            if userData in default_value:
                item_widget.setSelected(True)
        layout.addWidget(listwidget)
        return panel, self._get_multi_selection_value

    def _get_multi_selection_value(self):
        listwidget = self._ui.inputWidget
        return [item.data(Qt.Qt.UserRole) for item in listwidget.selectedItems()]

    def _create_group_panel(self, input_data):
        title = input_data.get('key', '')
        unit = input_data.get('unit', '')
        if unit:
            title += "(" + unit + ")"
        panel = Qt.QGroupBox(title)
        layout = Qt.QVBoxLayout()
        panel.setLayout(layout)
        self._ui.inputLabel = Qt.QLabel()
        self._ui.unitLabel = Qt.QLabel()
        return panel

    def _create_simple_panel(self, input_data):
        panel = Qt.QWidget()
        key = input_data.get('key', '')
        if key:
            key += ":"
        unit = input_data.get('unit', '')
        layout = Qt.QGridLayout()
        panel.setLayout(layout)
        self._ui.inputLabel = label = Qt.QLabel(key)
        layout.addWidget(label, 0, 0)
        self._ui.unitLabel = unit = Qt.QLabel(unit)
        layout.addWidget(unit, 0, 2)
        layout.setColumnStretch(0, 0)
        layout.setColumnStretch(1, 1)
        layout.setColumnStretch(2, 0)
        return panel

    def create_integer_panel(self, input_data):
        panel = self._create_simple_panel(input_data)
        minimum = input_data.get('minimum', numpy.iinfo('i').min)
        maximum = input_data.get('maximum', numpy.iinfo('i').max)
        step = input_data.get('step', 1)
        layout = panel.layout()
        self._ui.inputWidget = spinbox = Qt.QSpinBox()
        spinbox.setMinimum(minimum)
        spinbox.setMaximum(maximum)
        spinbox.setSingleStep(step)
        if 'default_value' in input_data:
            spinbox.setValue(input_data['default_value'])
        layout.addWidget(spinbox, 0, 1)
        return panel, self._get_integer_value

    def _get_integer_value(self):
        return int(self._ui.inputWidget.value())

    def create_float_panel(self, input_data):
        panel = self._create_simple_panel(input_data)
        minimum = input_data.get('minimum', numpy.finfo('d').min)
        maximum = input_data.get('maximum', numpy.finfo('d').max)
        step = input_data.get('step', 1)
        decimals = input_data.get('decimals', 1)
        layout = panel.layout()
        self._ui.inputWidget = spinbox = Qt.QDoubleSpinBox()
        spinbox.setMinimum(minimum)
        spinbox.setMaximum(maximum)
        spinbox.setSingleStep(step)
        spinbox.setDecimals(decimals)
        if 'default_value' in input_data:
            spinbox.setValue(input_data['default_value'])
        layout.addWidget(spinbox, 0, 1)
        return panel, self._get_float_value

    def _get_float_value(self):
        return float(self._ui.inputWidget.value())

    def create_string_panel(self, input_data):
        panel = self._create_simple_panel(input_data)
        layout = panel.layout()
        self._ui.inputWidget = lineedit = Qt.QLineEdit()
        if 'default_value' in input_data:
            lineedit.setText(str(input_data['default_value']))
        lineedit.selectAll()
        layout.addWidget(lineedit, 0, 1)
        return panel, self._get_string_value

    def _get_string_value(self):
        return str(self._ui.inputWidget.text())

    def create_text_panel(self, input_data):
        panel = self._create_group_panel(input_data)
        layout = panel.layout()
        self._ui.inputWidget = textedit = Qt.QTextEdit()
        textedit.setAcceptRichText(False)
        if 'default_value' in input_data:
            textedit.setPlainText(str(input_data['default_value']))
        textedit.selectAll()
        layout.addWidget(textedit)
        return panel, self._get_text_value

    def _get_text_value(self):
        return str(self._ui.inputWidget.toPlainText())

    def create_boolean_panel(self, input_data):
        panel = self._create_simple_panel(input_data)
        layout = panel.layout()
        self._ui.inputWidget = checkbox = Qt.QCheckBox()
        value = input_data.get('default_value', False)
        checkbox.setChecked(value)
        layout.addWidget(checkbox, 0, 1)
        return panel, self._get_boolean_value

    def _get_boolean_value(self):
        return self._ui.inputWidget.checkState() == Qt.Qt.Checked

    def inputPanel(self):
        return self._ui._inputPanel

    def buttonBox(self):
        """Returns the button box from this panel

        :return: the button box from this panel
        :rtype: PyQt4.Qt.QDialogButtonBox"""
        return self._ui._buttonBox

    def addButton(self, button, role=Qt.QDialogButtonBox.ActionRole):
        """Adds the given button with the given to the button box

        :param button: the button to be added
        :type button: PyQt4.QtGui.QPushButton
        :param role: button role
        :type role: PyQt4.Qt.QDialogButtonBox.ButtonRole"""
        self._ui._buttonBox.addButton(button, role)

    def setIconPixmap(self, pixmap):
        """Sets the icon to the dialog

        :param pixmap: the icon pixmap
        :type pixmap: PyQt4.Qt.QPixmap"""
        self._ui.iconLabel.setPixmap(pixmap)

    def setText(self, text):
        """Sets the text of this panel

        :param text: the new text
        :type text: str"""
        self._ui.textLabel.setText(text)

    def getText(self):
        """Returns the current text of this panel

        :return: the text for this panel
        :rtype: str"""
        return self._ui.textLabel.text()

    def setInputFocus(self):
        inputWidget = self._ui.inputWidget
        if not inputWidget:
            return
        if isinstance(inputWidget, Qt.QWidget):
            inputWidget.setFocus()
        elif isinstance(inputWidget, Qt.QButtonGroup):
            bid = inputWidget.checkedId()
            if bid < 0:
                button = inputWidget.buttons()[0]
            else:
                button = inputWidget.button(bid)
            button.setFocus()


def main():
    app = Qt.QApplication([])

    class Listener(object):

        def on_accept(self):
            print("user selected", self.panel.value())

    d = dict(prompt="What's your favourite car brand?",
             data_type=["Mazda", "Skoda", "Citroen",
                        "Mercedes", "Audi", "Ferrari"],
             default_value="Mercedes")
    w = TaurusInputPanel(d)
    l = Listener()
    l.panel = w
    w.buttonBox().accepted.connect(l.on_accept)
    w.show()
    app.exec_()

if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
Taurus Widget qwt5 (former `plot`) module
==========================================


This module provides the Qwt5-based widgets previously provided by
`taurus.qt.qtgui.plot`.
Note that `PyQwt5 module <http://pyqwt.sourceforge.net/>`_
only works with Python2 and PyQt4 and is no longer supported,
so taurus is moving to other modules for plotting (pyqtgraph, silx, ...)
"""

from .qwtdialog import TaurusPlotConfigDialog
from .scales import *
from .taurusplot import *
from .taurustrend import *
from .arrayedit import ArrayEditor
from .taurusarrayedit import TaurusArrayEditor
from .curvesAppearanceChooserDlg import CurveAppearanceProperties, CurvesAppearanceChooser
from .curveprops import CurvePropertiesView
from .monitor import TaurusMonitorTiny
from .curveStatsDlg import CurveStatsDialog



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Source code text utilities"""

__all__ = ["get_eol_chars", "get_os_name_from_eol_chars",
           "get_eol_chars_from_os_name", "has_mixed_eol_chars",
           "fix_indentation"]

__docformat__ = 'restructuredtext'

# Order is important:
EOL_CHARS = (("\r\n", 'nt'), ("\n", 'posix'), ("\r", 'mac'))


def get_eol_chars(text):
    """Get text EOL characters"""
    for eol_chars, _os_name in EOL_CHARS:
        if text.find(eol_chars) > -1:
            return eol_chars


def get_os_name_from_eol_chars(eol_chars):
    """Return OS name from EOL characters"""
    for chars, os_name in EOL_CHARS:
        if eol_chars == chars:
            return os_name


def get_eol_chars_from_os_name(os_name):
    """Return EOL characters from OS name"""
    for eol_chars, name in EOL_CHARS:
        if name == os_name:
            return eol_chars


def has_mixed_eol_chars(text):
    """Detect if text has mixed EOL characters"""
    eol_chars = get_eol_chars(text)
    if eol_chars is None:
        return False
    correct_text = eol_chars.join((text + eol_chars).splitlines())
    return repr(correct_text) != repr(text)


def fix_indentation(text):
    """Replace tabs by spaces"""
    return text.replace('\t', ' ' * 4)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################


from __future__ import absolute_import
from taurus.external.qt import Qt, compat
import taurus
import numpy

from taurus.qt.qtgui.container import TaurusWidget
from .arrayedit import ArrayEditor
from functools import partial

class TaurusArrayEditor(TaurusWidget):

    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent=parent, designMode=designMode)

        self._xAttr = None
        self._yAttr = None

        self._arrayEditor = ArrayEditor()
        self.fromFileBT = Qt.QPushButton('Read From File')
        self.toFileBT = Qt.QPushButton('Write To File')
        self.fromAttrBT = Qt.QPushButton('Read From Device')
        self.toAttrBT = Qt.QPushButton('Write To Device')

        self.fromAttrBT.setEnabled(False)
        self.toAttrBT.setEnabled(False)

        layout = Qt.QGridLayout(self)
        layout.addWidget(self._arrayEditor, 0, 0, 1, 4)
        layout.addWidget(self.fromFileBT, 1, 0)
        layout.addWidget(self.toFileBT, 1, 1)
        layout.addWidget(self.fromAttrBT, 1, 2)
        layout.addWidget(self.toAttrBT, 1, 3)

        self.fromFileBT.clicked.connect(self._onFromFile)
        self.toFileBT.clicked.connect(self.onToFile)
        self.fromAttrBT.clicked.connect(partial(self.onFromAttr, quiet=False))
        self.toAttrBT.clicked.connect(partial(self.onToAttr, quiet=False))

    def arrayEditor(self):
        return self._arrayEditor

    def setModel(self, model):
        '''returns True if a curve could be set from the attribute. Flase otherwise'''
        if not model:
            self._xAttr = self._yAttr = None
            self.fromAttrBT.setEnabled(False)
            return
        attrs = str(model).split("|")
        self._yAttr = taurus.Attribute(attrs[-1])
        if self._yAttr is None or len(attrs) not in (1, 2):
            self.error('Invalid model for %s' % str(self.__class__))
            self.fromAttrBT.setEnabled(False)
            return
        if len(attrs) == 1:
            self._xAttr = None
        else:
            self._xAttr = taurus.Attribute(attrs[0])

        # set the master curve from the attribute(s)
        ok = self.onFromAttr(quiet=True)

        # enable/disable the attribute I/O buttons
        self.fromAttrBT.setEnabled(True)
        enableWrite = (self._xAttr is None or self._xAttr.isWritable()
                       ) and self._yAttr.isWritable()
        self.toAttrBT.setEnabled(True)

        return ok

    def _onFromFile(self):
        """dummy, just to be used as an unambiguous slot"""
        self.onFromFile()

    def onFromFile(self, filename=None, **kwargs):
        '''imports Master curve from a two-column ASCII file.
        The first colum will be interpreted to be the abcissas.
        If filename is not given, a dialog for choosing a file is presented
        kwargs can contain keyword arguments to pass to numpy.loadtxt() when reading each file
        accepted keywords and their default values are:
        {dtype=<type 'float'>, comments='#', delimiter=None, converters=None,
        skiprows=0, usecols=None, unpack=False}
        see help from numpy.loadtxt for more info on the kwargs'''
        if filename is None:
            filename, _ = compat.getOpenFileName(
                self, 'Choose input file', '', 'Ascii file (*)')
        if not filename:
            return False
        filename = str(filename)
        try:
            M = numpy.loadtxt(filename, **kwargs)
            if len(M.shape) != 2:
                raise Exception()
            x = M[:, 0]
            y = M[:, 1]
        except:
            self.error('Invalid input file "%s"' % filename)
            Qt.QMessageBox.warning(
                self, 'Invalid input file', 'Unknown format in selected file.\n Hint: The file must be two-column ASCII')
            return
        self._arrayEditor.setMaster(x, y)

    def onToFile(self):
        '''writes the Corrected curve to an ascii file'''
        if self._arrayEditor.plot1.exportAscii(curves=['Corrected']):
            x, y = self._arrayEditor.getCorrected()
            self._arrayEditor.setMaster(x, y, keepCP=True)

    def onFromAttr(self, quiet=False):
        '''reads the Master curve from the attributes set by model.
        '''
        if self._yAttr is None:
            return False  # @TODO: launch a dialog for choosing the x and y attributes
        try:
            y = numpy.array(self._yAttr.read().rvalue)
            if self._xAttr is None:
                x = numpy.arange(y.size)
            else:
                x = numpy.array(self._xAttr.read().rvalue)
        except Exception as e:
            self.error('Error reading from attribute(s): %s' % (str(e)))
            if not quiet:
                Qt.QMessageBox.warning(
                    self, 'Error Reading Attribute', 'Cannot read master curve')
            return False

        if quiet:
            should_set = True
        else:
            answer = Qt.QMessageBox.question(
                    self, 'Read from attributes?',
                    'Read Master curve from attributes?',
                    Qt.QMessageBox.Yes|Qt.QMessageBox.No
            )
            should_set = (answer == Qt.QMessageBox.Yes == answer)

        if should_set:
            try:
                self._arrayEditor.setMaster(x, y)
            except ValueError:
                self.error('Cannot set master curve from attributes')
                if not quiet:
                    Qt.QMessageBox.warning(self, 'Error',
                            'Cannot set master curve from attributes')
                return False
        return True

    def onToAttr(self, quiet=False):
        '''writes the Corrected curve to the attributes set by the model'''
        if self._yAttr is None:
            return  # @TODO: launch a dialog for choosing the x and y attributes
        x, y = self._arrayEditor.getCorrected()
        try:
            self._yAttr.write(y)
            if self._xAttr is not None:
                self._xAttr.write(x)
            # checking that the write was actually ok (this should be changed
            # when taurus provides a mechanism for this)
            if numpy.any(self._yAttr.read(cache=False).wvalue != y):
                raise IOError('Unexpected Write error: %s' %
                              self._yAttr.getFullName())
            if self._xAttr is not None and numpy.any(self._xAttr.read(cache=False).wvalue != x):
                raise IOError('Unexpected Write error: %s' %
                              self._xAttr.getFullName())
        except Exception as e:
            self.error('Error writing to attribute(s): %s' % (str(e)))
            if not quiet:
                Qt.QMessageBox.warning(
                    self, 'Error Writing attribute', 'Cannot write corrected curve to attribute')
            return
        self._arrayEditor.setMaster(x, y, keepCP=True)
        if not quiet:
            Qt.QMessageBox.information(
                self, 'Success', 'Corrected curve has been written')

    @classmethod
    def getQtDesignerPluginInfo(cls):
        """Returns pertinent information in order to be able to build a valid
        QtDesigner widget plugin

        :return: (dict) a map with pertinent designer information"""
        ret = TaurusWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.qwt5'
        ret['group'] = 'Taurus Input'
        ret['icon'] = 'designer:arrayedit.png'
        ret['container'] = False
        return ret



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains a taurus qt style called nebula"""

__all__ = ["getStyle", "getStyleSheet"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt

_NEBULA_KEYS = {
    'border_radius': '4px',
    'titlebar_background_color': 'qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 rgb(200, 200, 200), stop: 1 rgb(150, 150, 150))',
    'selected_titlebar_background_color': 'qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 rgb(60, 150, 255), stop: 1 rgb(0, 65, 200))',
    'single_titlebar_background_color': 'qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 rgb(90, 180, 255), stop: 1 rgb(30, 95, 250))',
    'titlebar_color': 'white',
    'selected_titlebar_color': 'white',
    'content_background_color': 'qlineargradient(x1: 0, y1: 0, x2: 1.0, y2: 1.0, stop: 0 rgb(224, 224, 224), stop: 1 rgb(255, 255, 255))'
}

_NEBULA_STYLESHEET = \
    """QToolBox:tab {{
    color: {titlebar_color};
    border-width: 0px;
    border-style: solid;
    border-color: rgb(0, 65, 200);
    border-top-left-radius: 0px;
    border-top-right-radius: {border_radius};
    border-bottom-left-radius: {border_radius};
    border-bottom-right-radius: {border_radius};
    background-color: {titlebar_background_color};
}}

QToolBox:tab:selected {{
    background-color: {selected_titlebar_background_color};
}}

QToolBox:tab:first {{
    border-top-left-radius: 0px;
    border-top-right-radius: 0px;
}}

QToolBox:tab:last {{
    border-bottom-left-radius: 0px;
    border-bottom-right-radius: 0px;
}}

QToolBox:tab:only-one {{
    background-color: {single_titlebar_background_color};
}}

QDockWidget {{
    color: {titlebar_color};
    background-color: {content_background_color};
    titlebar-close-icon: url(:/titlebar_close_black.png);
    titlebar-normal-icon: url(:/titlebar_undock_black.png);
}}

QDockWidget::title {{
    text-align: left;
    padding-left: {border_radius};

    border-top-left-radius: {border_radius};
    border-top-right-radius: {border_radius};
    border-bottom-left-radius: {border_radius};
    border-bottom-right-radius: {border_radius};

    background-color: {selected_titlebar_background_color};
}}

 QGroupBox {{
    border: 1px solid;
    border-color: rgb(0, 65, 200);
    border-radius: {border_radius};
    margin-top: 1.5ex;
    padding-top: 8px;
    background-color: {content_background_color};
 }}

QGroupBox::title {{
    subcontrol-origin: margin;
    subcontrol-position: top left;
    padding-top: 1px;
    padding-right: 3px;
    padding-bottom: 2px;
    padding-left: 3px;
    border-width: 0px;
    border-radius: {border_radius};
    color:white;
    background-color: {selected_titlebar_background_color};
    left: 5px;
}}

 QGroupBox::indicator {{
    width: 15px;
    height: 15px;
}}

QTabWidget {{

}}

QTabWidget::tab-bar {{
    left: 6px;
}}

QTabWidget::pane {{
    border: 1px solid;
    border-color: rgb(0, 65, 200);
    border-top-left-radius: {border_radius};
    border-top-right-radius: {border_radius};
    border-bottom-left-radius: {border_radius};
    border-bottom-right-radius: {border_radius};
    background-color: {content_background_color};
}}

QTabBar::tab {{
    color:white;
    border-bottom-color: rgb(0, 65, 200);
    background-color: {titlebar_background_color};
    min-width: 8ex;
    padding: 2px;
}}

QTabBar::tab:top {{
    border-top-left-radius: {border_radius};
    border-top-right-radius: {border_radius};
}}

QTabBar::tab:bottom {{
    border-bottom-left-radius: {border_radius};
    border-bottom-right-radius: {border_radius};
}}

QTabBar::tab:selected {{
    background-color: {selected_titlebar_background_color};
}}


/*
 QMainWindow::separator {{
    background: yellow;
    width: 2px;
    height: 2px;
 }}

 QMainWindow::separator:hover {{
    background: red;
 }}

 */


"""

NEBULA_STYLESHEET = _NEBULA_STYLESHEET.format(**_NEBULA_KEYS)


def getStyle():
    return None


def getStyleSheet():
    return NEBULA_STYLESHEET



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Enumeration module.
In C, enums allow you to declare a bunch of constants with unique values,
without necessarily specifying the actual values (except in cases where you
need to). Python has an accepted idiom that's fine for very small numbers of
constants (A, B, C, D = range(4)) but it doesn't scale well to large numbers,
and it doesn't allow you to specify values for some constants while leaving
others unspecified. This approach does those things, while verifying that all
values (specified and unspecified) are unique. Enum values then are attributes
of an Enumeration class (Volkswagen.BEETLE, Volkswagen.PASSAT, etc.)."""

from builtins import int
from builtins import str
from builtins import object
from future.utils import string_types


__all__ = ["EnumException", "Enumeration"]
__docformat__ = "restructuredtext"


class EnumException(Exception):
    """Exception thrown by :class:`Enumeration` when trying to declare an
    invalid enumeration."""
    pass


class Enumeration(object):
    """ Enumeration class intended to provide the 'enum' feature present in many
    programming languages.
    The elements of the enumeration can be accessed in an "object member way" or
    as elements of a dictionary.
    Usage::

        from taurus.core.util.enumeration import Enumeration

        Volkswagen = Enumeration("Volkswagen",
            ["JETTA",
             "RABBIT",
             "BEETLE",
             ("THING", 400),
             "PASSAT",
             "GOLF",
             ("CABRIO", 700),
             "EURO_VAN",
             "CLASSIC_BEETLE",
             "CLASSIC_VAN"
             ])

    In the command line::

        >>> my_car = Volkswagen.BEETLE
        >>> homer_car = Volkswagen.PASSAT

        >>> print Volkswagen.BEETLE
        2

        >>> print Volkswagen['BEETLE']
        2

        >>>print Volkswagen.whatis(homer_car)
        'PASSAT'
    """

    def __init__(self, name, enumList, flaggable=False, no_doc=False):
        self._name = name
        lookup = {}
        reverseLookup = {}
        uniqueNames = set()
        self._flaggable = flaggable
        self._uniqueValues = uniqueValues = set()
        self._uniqueId = 0
        for x in enumList:
            if isinstance(x, tuple):
                if flaggable:
                    raise EnumException(
                        "flagable enum does not accept tuple items")
                x, i = x
                if not isinstance(x, string_types):
                    raise EnumException("enum name is not a string: " + str(x))
                if not isinstance(i, int):
                    raise EnumException(
                        "enum value is not an integer: " + str(i))
                if x in uniqueNames:
                    raise EnumException("enum name is not unique: " + str(x))
                if i in uniqueValues:
                    raise EnumException(
                        "enum value is not unique for " + str(x))
                uniqueNames.add(x)
                uniqueValues.add(i)
                lookup[x] = i
                reverseLookup[i] = x
        for x in enumList:
            if not isinstance(x, tuple):
                if not isinstance(x, string_types):
                    raise EnumException("enum name is not a string: " + str(x))
                if x in uniqueNames:
                    raise EnumException("enum name is not unique: " + str(x))
                uniqueNames.add(x)
                i = self._generateUniqueId()
                uniqueValues.add(i)
                lookup[x] = i
                reverseLookup[i] = x
        self.lookup = lookup
        self.reverseLookup = reverseLookup
        if not no_doc:
            self.__doc_enum()

    def __call__(self, i):
        # TODO: Dummy implementation to simulate Python Enum behaviour.
        # It is not a complete replacement because although we can use
        # Enumeration as Callable, it still return an int instead of an
        # Enumeration member.
        return self.lookup[self.whatis(i)]

    def __len__(self):
        return len(self.lookup)

    def _generateUniqueId(self):
        if self._flaggable:
            n = 2 ** self._uniqueId
        else:
            while self._uniqueId in self._uniqueValues:
                self._uniqueId += 1
            n = self._uniqueId
        self._uniqueId += 1
        return n

    def __contains__(self, i):
        if isinstance(i, int):
            return i in self.reverseLookup
        elif isinstance(i, string_types):
            return i in self.lookup

    def __getitem__(self, i):
        if isinstance(i, int):
            return self.whatis(i)
        elif isinstance(i, string_types):
            return self.lookup[i]

    def __getattr__(self, attr):
        if attr not in self:
            raise AttributeError
        return self.lookup[attr]

    def __doc_enum(self):
        rl = self.reverseLookup
        keys = list(rl)
        keys.sort()
        values = "\n".join(["    - {0} ({1})".format(rl[k], k) for k in keys])
        self.__doc__ = self._name + " enumeration. " + \
            "Possible values are:\n\n" + values

    def __str__(self):
        rl = self.reverseLookup
        keys = list(rl)
        keys.sort()
        values = ", ".join([rl[k] for k in keys])
        return self._name + "(" + values + ")"

    def __repr__(self):
        rl = self.reverseLookup
        keys = list(rl)
        keys.sort()
        values = [rl[k] for k in keys]
        return "Enumeration('" + self._name + "', " + str(values) + ")"

    def has_key(self, key):
        """Determines if the enumeration contains the given key
        :param key: the key
        :type  key: str
        :return: True if the key is in the enumeration or False otherswise
        :rtype: bool"""
        return key in self.lookup

    def keys(self):
        """Returns an iterable containning the valid enumeration keys
        :return: an interable containning the valid enumeration keys
        :rtype: iter<str>"""
        return list(self.lookup.keys())

    def whatis(self, value):
        """Returns a string representation of the value in the enumeration.
        :param value: a valid enumeration element
        :return: a string representation of the given enumeration element
        :rtype: str"""
        return self.reverseLookup[value]

    def get(self, i):
        """Returns the element for the given key/value"""
        return self[i]
